/* dns.h of iplog v1.5 -- eric@ojnk.org */
#include <netdb.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <netdb.h>

#define DNS_TIME	1800
#define DNS_MAXSIZE	32
#define DNS_HASH(x)	(((x >> 24) + (x >> 16)) % (DNS_MAXSIZE + 1))

int hits = 0, total = 0;

void sigunused(int stat)
{
	if (total)
		syslog(LEVEL, "%d/%d cached; %3.2f%% success rate", hits, total, ((float)hits / (float)total) * 100);
	else
		syslog(LEVEL, "no queries have been made yet");
	signal(SIGUNUSED, sigunused);
}

struct dns_cache {
	unsigned long address;
	char *host;
	time_t when;
};

typedef struct dns_cache dns_entry;
dns_entry dns_table[DNS_MAXSIZE];

void init_dns_cache(void)
{
	int n;

	signal(SIGUNUSED, sigunused);

	for (n = 0; n < DNS_MAXSIZE; n++) {
		if (sizeof(*dns_table[n].host) > 1)
			free(dns_table[n].host);
		dns_table[n].address = dns_table[n].when = (-1L);
	}
}

char *get_dns_cache(u_long addr)
{
	int n = DNS_HASH(addr);

	total++;

	if (dns_table[n].address == addr) {
		hits++;
		return (dns_table[n].host);
	}
	add_dns_cache(addr);

	if (dns_table[n].address == addr)
		return (dns_table[n].host);
	else
		return NULL;
}

void add_dns_cache(u_long addr)
{
	struct in_addr in;
	struct hostent *he;
	int n = DNS_HASH(addr);

	in.s_addr = addr;
	he = gethostbyaddr((char *)&in, sizeof(struct in_addr), AF_INET);

	if (he != NULL) {
		dns_table[n].address = addr;
		dns_table[n].host = strdup(he->h_name);
		time(&dns_table[n].when);
	}
}

void expire_dns()
{
	time_t tm;
	int n;

	time(&tm);

	for (n = 0; n < DNS_MAXSIZE; n++)
		if (dns_table[n].address != (-1L))
			if (tm >= (dns_table[n].when + DNS_TIME)) {
				if (sizeof(*dns_table[n].host) > 1)
					free(dns_table[n].host);
				dns_table[n].address = dns_table[n].when = (-1L);
			}
}
