/* smurf.h of iplog v1.5 -- eric@ojnk.org */
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <netdb.h>

#define HDR_SIZE sizeof(struct ip)

#define SMURF_WAIT	60

#define THRESH_PPS	25			/* packets per second */
#define THRESH_KBPS	20			/* kilobytes per second */

time_t last = 0, lastalarm = 0;
float kbps = 0.0;
int count = 0, smurf_mode = 0, pps = 0;
unsigned long int *ips;

void sigalrm(int stat)
{
	if (smurf_mode) {
		if (sizeof(*ips) > 1)
			free(ips);
		smurf_mode = 0;
		syslog(LEVEL, "No longer in smurf mode");
	}
	signal(SIGALRM, sigalrm);
}

int comp(const void *a, const void *b)
{
	return *(int *)a - *(int *)b;
}

/* based very heavily on buglord's SmurfLog, but modified to work in iplog */
int smurf(struct packet pkt)
{
	time_t current;
	u_char *p;
	unsigned long int the_ip;

	time(&current);

	if (smurf_mode) {
		p = (unsigned char *)&pkt.ip.ip_src.s_addr;

		p[3] = 0;

		the_ip = pkt.ip.ip_src.s_addr;

		if (bsearch(&the_ip, ips, count, sizeof(*ips), comp) == NULL) {
			syslog(LEVEL, "smurf attack detected from %s/24 (%d bytes)", inet_ntoa(pkt.ip.ip_src), UNFIX(pkt.ip.ip_len) - HDR_SIZE);
			if ((ips = realloc(ips, count + 1)) == NULL)
				fatal("smurf prot realloc: %m");
			ips[count++] = the_ip;
			qsort(ips, count, sizeof(*ips), comp);
		}
		last = current;
		if ((current >= (lastalarm + 5))) {
			alarm(SMURF_WAIT);
			lastalarm = current;
		}
		return (0);
	} else {
		kbps += UNFIX(pkt.ip.ip_len) / 1024.0;
		pps++;

		if ((current - last) >= 1 || (current < last)) {
			pps = 0;
			kbps = 0.0;
			last = current;
			return (1);
		} else if (pps >= THRESH_PPS && kbps >= THRESH_KBPS) {
			smurf_mode = 1;
			syslog(LEVEL, "threshold met (%ld pkt/s, %4.02fkb/s)", pps, kbps);
			last = current;
			return (0);
		}
		last = current;
		return (1);
	}
}
