/* udplog.c of iplog v1.5 -- eric@ojnk.org */
#include <unistd.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <fcntl.h>
#include <ctype.h>
#include <arpa/inet.h>

#include "include/iplog.h"
#include "include/lockfile.h"

#define UDP_SIZE		(sizeof(struct ip)+sizeof(struct udphdr))
#define MAX_NAMESERVERS 64		/* i think that's reasonable.. */

#ifndef _PATH_RESCONF			/* bleh */
#define _PATH_RESCONF	"/etc/resolv.conf"
#endif

struct packet {
	struct ip ip;
	struct udphdr udp;
};

#ifdef DETECT_TRACEROUTE
#define dp	UNFIX(pkt.udp.dest)
#define sp	UNFIX(pkt.udp.source)
#define ttl pkt.ip.ip_ttl

static char *hostlookup(u_long);

#endif
static char *servlookup(u_short);
void reread_ns(int);
void readns(u_long *);
u_long name_resolve(char *);
u_long ignore[MAX_NAMESERVERS];

int main(void)
{
	struct packet pkt;
	struct in_addr in;
	int sock, i, bad = 0;

#ifdef DETECT_TRACEROUTE
	time_t lasttr = 0;
	u_long last = 0;

#endif

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) < 0) {
		perror("socket");
		exit(-1);
	}
	openlog("udplog", LOGGING_OPTIONS, LOGGING_FACILITY);
	signal(SIGHUP, reread_ns);

	fork_to_background();
	initialize_lockfile("udplog");
	drop_privileges();
	readns(ignore);

	for (;;) {
		read(sock, (struct header *)&pkt, sizeof(pkt));

		if (UNFIX(pkt.udp.source) == 26000 || UNFIX(pkt.udp.source) == 27500 || UNFIX(pkt.udp.source) == 27501)		/* Quake */
			continue;

		if (UNFIX(pkt.udp.source) == 53) {
			for (i = 0, bad = 0; ignore[i] != (-1L); i++) {
				if (pkt.ip.ip_src.s_addr == ignore[i]) {
					bad = 1;
					break;
				}
			}
		} else
			bad = 0;

		if (!bad) {
			in.s_addr = pkt.ip.ip_src.s_addr;

#ifdef DETECT_TRACEROUTE
			if (((dp > 32000) && (dp < 40000)) && (sp > 30000) && (ttl == 1)) {
				if ((last != in.s_addr) || (time(NULL) >= (lasttr + 15))) {
					syslog(LEVEL, "traceroute from %s (%s)", hostlookup(in.s_addr), inet_ntoa(in));
					time(&lasttr);
					last = in.s_addr;
				}
			} else
#endif
				syslog(LEVEL, "dgram to %s from %s:%u sz=%u(+%u)", servlookup(pkt.udp.dest), inet_ntoa(in), UNFIX(pkt.udp.source), UNFIX(pkt.ip.ip_len) - UDP_SIZE, UDP_SIZE);
		}
		memset((char *)&pkt, 0, sizeof(pkt));
	}
	close(sock);
}

static char *servlookup(u_short port)
{
	struct servent *se;

	if ((se = getservbyport(port, "udp")) != NULL)
		return (se->s_name);
	else {
		static char buff[16];

		sprintf(buff, "port %u", UNFIX(port));
		return (buff);
	}
}

#ifdef DETECT_TRACEROUTE
static char *hostlookup(u_long addr)
{
	register struct hostent *he;
	struct in_addr i;

	i.s_addr = addr;
	he = gethostbyaddr((char *)&i, sizeof(struct in_addr), AF_INET);

	if (he != NULL)
		return (he->h_name);
	else
		return (inet_ntoa(i));
}
#endif

void reread_ns(int stat)
{
	syslog(LEVEL, "caught hangup signal -- re-reading nameservers");
	readns(ignore);
	signal(SIGHUP, reread_ns);
}

void readns(u_long * ignore)
{
	FILE *fp;
	int found = 0;
	char line[BUFSIZ], host[18];

	if ((fp = fopen(_PATH_RESCONF, "r")) == NULL)
		fatal("%s: %m", _PATH_RESCONF);

	while ((fgets(line, sizeof(line) - 1, fp) != NULL) && (found != MAX_NAMESERVERS))
		if (sscanf(line, "nameserver %18s", host) == 1)
			if ((ignore[found++] = name_resolve(host)) == (-1L))
				found--;

	(void)fclose(fp);

	syslog(LEVEL, "cached %d nameserver%s", found, (found == 1 ? "" : "s"));
	ignore[found++] = (-1L);

	return;
}

/* name_resolve function from route's libnet */
u_long name_resolve(char *host_name)
{
	u_long l;
	u_int val;
	int i;

	if (!isdigit(host_name[0]))
		return (-1L);

	l = 0;
	for (i = 0; i < 4; i++) {
		l <<= 8;
		if (*host_name) {
			val = 0;

			while (*host_name && *host_name != '.') {
				val *= 10;
				val += *host_name - '0';
				host_name++;
			}

			l |= val;
			if (*host_name)
				host_name++;
		}
	}
	return (htonl(l));
}
